// ---------------------------------------------------------------------------
/*
 * (c) Copyright 2010, Allan Petersen
 * ALL RIGHTS RESERVED
 * Permission to use, copy, modify, and distribute this software for
 * any purpose and without fee is hereby granted, provided that the above
 * copyright notice appear in all copies and that both the copyright notice
 * and this permission notice appear in supporting documentation, and that
 * the name of Allan Petersen not be used in advertising
 * or publicity pertaining to distribution of the software without specific,
 * written prior permission.
 *
 * THE MATERIAL EMBODIED ON THIS SOFTWARE IS PROVIDED TO YOU "AS-IS"
 * AND WITHOUT WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR OTHERWISE,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF MERCHANTABILITY OR
 * FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL ALLAN
 * PETERSEN BE LIABLE TO YOU OR ANYONE ELSE FOR ANY DIRECT,
 * SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY
 * KIND, OR ANY DAMAGES WHATSOEVER, INCLUDING WITHOUT LIMITATION,
 * LOSS OF PROFIT, LOSS OF USE, SAVINGS OR REVENUE, OR THE CLAIMS OF
 * THIRD PARTIES, WHETHER OR NOT ALLAN PETERSEN HAS BEEN
 * ADVISED OF THE POSSIBILITY OF SUCH LOSS, HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE
 * POSSESSION, USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * www.allanpetersen.com
 *
 */
// ---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include "GLee.h"
#include "Unit2.h"
// ---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma link "TOpenGLAPPanel"
#pragma resource "*.dfm"
TForm2 *Form2;

// ---------------------------------------------------------------------------
__fastcall TForm2::TForm2(TComponent* Owner) : TForm(Owner)
{
	rtri = 0;
	rquad = 0;
	Bitmap = new Graphics::TBitmap;
	FOnPaint = NULL;

	QueryPerformanceFrequency(&iFrequency);
	// ---- Initialize OpenGL
	OpenGLAPPanel1Init(NULL);
}

// ===========================================================================
///
///	Render to FBO
///
// ---------------------------------------------------------------------------
// 20-04-2010
// ===========================================================================
bool __fastcall TForm2::OffScreenRender()
{
	GLuint fb, color_rb, depth_rb;

	QueryPerformanceCounter(&StartTime);

	// RGBA8 RenderBuffer, 24 bit depth RenderBuffer, 256x256
	glGenFramebuffersEXT(1, &fb);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb);
	// Create and attach a color buffer
	glGenRenderbuffersEXT(1, &color_rb);
	// We must bind color_rb before we call glRenderbufferStorageEXT
	glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, color_rb);
	// The storage format is RGBA8
	glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_RGBA8,
		OpenGLAPPanel1->ClientWidth, OpenGLAPPanel1->ClientHeight);
	// Attach color buffer to FBO
	glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT,
		GL_RENDERBUFFER_EXT, color_rb);
	// -------------------------
	glGenRenderbuffersEXT(1, &depth_rb);
	glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, depth_rb);
	glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT24,
		OpenGLAPPanel1->ClientWidth, OpenGLAPPanel1->ClientHeight);
	// -------------------------
	// Attach depth buffer to FBO
	glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT,
		GL_RENDERBUFFER_EXT, depth_rb);
	// -------------------------
	// Does the GPU support current FBO configuration?
	GLenum status;
	status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
	switch(status)
	{
	case GL_FRAMEBUFFER_COMPLETE_EXT:
		break;
	default:

		break;
	}
	// -------------------------
	// and now you can render to the FBO (also called RenderBuffer)
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb);

	Render();

	CopyToBitmap(Bitmap);

	// Bind 0, which means render to back buffer
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);

	// Delete resources
	glDeleteRenderbuffersEXT(1, &color_rb);
	glDeleteRenderbuffersEXT(1, &depth_rb);
	// Bind 0, which means render to back buffer, as a result, fb is unbound
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
	glDeleteFramebuffersEXT(1, &fb);

	// ---- Calculating Frames per Second..
	// ---- This is the maximum frame rate and not the real one
	// ---- because the real frame rate is determined by TimerPaint
	// ---- Timer paint is set to 16ms ~ 62 FPS
	QueryPerformanceCounter(&StopTime);
	Fps = 1. / (((float)StopTime.QuadPart - (float)StartTime.QuadPart)
		* 1.0 / iFrequency.QuadPart);

	if (FOnPaint) {
		FOnPaint(Bitmap);
	}
}

// ===========================================================================
///
///	Render scene
///
// ---------------------------------------------------------------------------
// 19-04-2010
// ===========================================================================
bool __fastcall TForm2::Render()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	// Clear screen and depth buffer
	glLoadIdentity(); // Reset the current modelview matrix

	glTranslatef(-1.5f, 0.0f, -6.0f); // Move left 1.5 units and into the screen 6.0
	glRotatef(rtri, 0.0f, 1.0f, 0.0f); // Rotate the triangle on the Y axis ( NEW )

	glBegin(GL_TRIANGLES); // Drawing using triangles
	glColor3f(1.0f, 0.0f, 0.0f); // Red
	glVertex3f(0.0f, 1.0f, 0.0f); // Top of triangle (front)
	glColor3f(0.0f, 1.0f, 0.0f); // Green
	glVertex3f(-1.0f, -1.0f, 1.0f); // Left of triangle (front)
	glColor3f(0.0f, 0.0f, 1.0f); // Blue
	glVertex3f(1.0f, -1.0f, 1.0f); // Right of triangle (front)

	glColor3f(1.0f, 0.0f, 0.0f); // Red
	glVertex3f(0.0f, 1.0f, 0.0f); // Top of triangle (right)
	glColor3f(0.0f, 0.0f, 1.0f); // Blue
	glVertex3f(1.0f, -1.0f, 1.0f); // Left of triangle (right)
	glColor3f(0.0f, 1.0f, 0.0f); // Green
	glVertex3f(1.0f, -1.0f, -1.0f); // Right of triangle (right)

	glColor3f(1.0f, 0.0f, 0.0f); // Red
	glVertex3f(0.0f, 1.0f, 0.0f); // Top of triangle (back)
	glColor3f(0.0f, 1.0f, 0.0f); // Green
	glVertex3f(1.0f, -1.0f, -1.0f); // Left of triangle (back)
	glColor3f(0.0f, 0.0f, 1.0f); // Blue
	glVertex3f(-1.0f, -1.0f, -1.0f); // Right of triangle (back)

	glColor3f(1.0f, 0.0f, 0.0f); // Red
	glVertex3f(0.0f, 1.0f, 0.0f); // Top of triangle (left)
	glColor3f(0.0f, 0.0f, 1.0f); // Blue
	glVertex3f(-1.0f, -1.0f, -1.0f); // Left of triangle (left)
	glColor3f(0.0f, 1.0f, 0.0f); // Green
	glVertex3f(-1.0f, -1.0f, 1.0f); // Right of triangle (left)
	glEnd(); // Finished Drawing The Triangle

	glLoadIdentity(); // Reset the current modelview matrix
	glTranslatef(1.5f, 0.0f, -6.0f); // Move right 1.5 units and into the screen 6.0
	glRotatef(rquad, 1.0f, 0.0f, 0.0f); // Rotate the quad on the X axis ( NEW )
	glColor3f(0.5f, 0.5f, 1.0f); // Set the color to blue one time only

	glBegin(GL_QUADS); // Draw a quad
	glColor3f(0.0f, 1.0f, 0.0f); // Set the color to green
	glVertex3f(1.0f, 1.0f, -1.0f); // Top right of the quad (top)
	glVertex3f(-1.0f, 1.0f, -1.0f); // Top left of the quad (top)
	glVertex3f(-1.0f, 1.0f, 1.0f); // Bottom left of the quad (top)
	glVertex3f(1.0f, 1.0f, 1.0f); // Bottom right of the quad (top)

	glColor3f(1.0f, 0.5f, 0.0f); // Set the color to orange
	glVertex3f(1.0f, -1.0f, 1.0f); // Top right of the quad (bottom)
	glVertex3f(-1.0f, -1.0f, 1.0f); // Top left of the quad (bottom)
	glVertex3f(-1.0f, -1.0f, -1.0f); // Bottom left of the quad (bottom)
	glVertex3f(1.0f, -1.0f, -1.0f); // Bottom right of the quad (bottom)

	glColor3f(1.0f, 0.0f, 0.0f); // Set the color to red
	glVertex3f(1.0f, 1.0f, 1.0f); // Top right of the quad (front)
	glVertex3f(-1.0f, 1.0f, 1.0f); // Top left of the quad (front)
	glVertex3f(-1.0f, -1.0f, 1.0f); // Bottom left of the quad (front)
	glVertex3f(1.0f, -1.0f, 1.0f); // Bottom right of the quad (front)

	glColor3f(1.0f, 1.0f, 0.0f); // Set the color to yellow
	glVertex3f(1.0f, -1.0f, -1.0f); // Bottom left of the quad (back)
	glVertex3f(-1.0f, -1.0f, -1.0f); // Bottom right of the quad (back)
	glVertex3f(-1.0f, 1.0f, -1.0f); // Top right of the quad (back)
	glVertex3f(1.0f, 1.0f, -1.0f); // Top left of the quad (back)

	glColor3f(0.0f, 0.0f, 1.0f); // Set the color to blue
	glVertex3f(-1.0f, 1.0f, 1.0f); // Top right of the quad (left)
	glVertex3f(-1.0f, 1.0f, -1.0f); // Top left of the quad (left)
	glVertex3f(-1.0f, -1.0f, -1.0f); // Bottom left of the quad (left)
	glVertex3f(-1.0f, -1.0f, 1.0f); // Bottom right of the quad (left)

	glColor3f(1.0f, 0.0f, 1.0f); // Set the color to violet
	glVertex3f(1.0f, 1.0f, -1.0f); // Top right of the quad (right)
	glVertex3f(1.0f, 1.0f, 1.0f); // Top left of the quad (right)
	glVertex3f(1.0f, -1.0f, 1.0f); // Bottom left of the quad (right)
	glVertex3f(1.0f, -1.0f, -1.0f); // Bottom right of the quad (right)
	glEnd();

	rtri += 1.2f; // Increase the rotation variable for the triangle ( NEW )
	rquad -= 1.15f; // Decrease the rotation variable for the quad     ( NEW )

	return true; // Done drawing the quad
}

// ===========================================================================
// Name	 	: CopyToBitmap
// Purpose	 	: Copy OpenGL Window to Bitmap.
// Return		:
// ---------------------------------------------------------------------------
// 2006-04-04	Created by AP.
// ===========================================================================
bool __fastcall TForm2::CopyToBitmap(Graphics::TBitmap *BitMap)
{
	// ---- Creating Directly into bitmap
	int Bytes, yy;
	SIZE glsize;
	unsigned char *pPixelDataStart;
	unsigned char *pPixelData;
	TRGBQuad *Ptr;
	// Byte      *Ptr;

	glsize.cx = OpenGLAPPanel1->ClientWidth;
	glsize.cy = OpenGLAPPanel1->ClientHeight;

	// ---- Lines have to be 32 bytes aligned, suppose 24 bits per pixel
	// ---- just cropped it
	glsize.cx -= glsize.cx % 4;
	glsize.cy -= glsize.cy % 4;

	// ---- Alloc pixel bytes
	Bytes = ((glsize.cx * 24 + 31) / 32 * 4) * glsize.cy;
	if (Bytes < 4 || glsize.cy < 1) {
		return(false);
	}

	pPixelData = new unsigned char[Bytes];

	// ---- Copy from OpenGL
	glReadPixels(0, 0, glsize.cx, glsize.cy, GL_BGR_EXT, GL_UNSIGNED_BYTE,
		pPixelData);
	/*
	int rc = glGetError ();
	if (rc != GL_NO_ERROR) {
	return (false);
	}
	 */

	// ---- Setting up bitmap, must be 32 bit to contain an alpha channel
	BitMap->PixelFormat = pf32bit;
	BitMap->HandleType = bmDIB;
	BitMap->Width = glsize.cx;
	BitMap->Height = glsize.cy;
	BitMap->AlphaFormat = afPremultiplied;

	// ---- Calculating bytes
	Bytes = (24 * glsize.cx) >> 3;
	while (Bytes % 4)
		Bytes++;

	pPixelDataStart = pPixelData;
	// ---- Vertical pixels
	for (yy = glsize.cy - 1; yy >= 0; yy--) {
		Ptr = reinterpret_cast<TRGBQuad*>(BitMap->ScanLine[yy]);

		// ---- Horizontal pixels to tranfer individual values and handle the alpha channel
		// ---- this can be made faster by using the alpha channel directly in OpenGL
		// ---- and then copy the buffer directly from OpenGL to Bitmap.
		for (int xx = 0; xx < glsize.cx; xx++) {
			Ptr[xx].rgbBlue = *pPixelData;
			pPixelData += sizeof(unsigned char);
			Ptr[xx].rgbGreen = *pPixelData;
			pPixelData += sizeof(unsigned char);
			Ptr[xx].rgbRed = *pPixelData;
			pPixelData += sizeof(unsigned char);
			Ptr[xx].rgbReserved = 255;
			if (Ptr[xx].rgbBlue == 0 && Ptr[xx].rgbGreen == 0 && Ptr[xx]
				.rgbRed == 0) {
				Ptr[xx].rgbReserved = 0; // ---- Make black fully transparent
			}
		}
	}
	delete[]pPixelDataStart;

	return(true);
}

// ---------------------------------------------------------------------------
void __fastcall TForm2::OpenGLAPPanel1Resize(TObject *Sender)
{
	if (OpenGLAPPanel1->Height == 0) // Prevent A Divide By Zero By
	{
		OpenGLAPPanel1->Height = 1; // Making height equal One
	}

	glViewport(0, 0, OpenGLAPPanel1->Width, OpenGLAPPanel1->Height);
	// Reset the current viewport

	glMatrixMode(GL_PROJECTION); // Select the projection matrix
	glLoadIdentity(); // Reset the projection matrix

	// Calculate the aspect ratio of the window
	gluPerspective(45.0f, (GLfloat)OpenGLAPPanel1->Width / (GLfloat) OpenGLAPPanel1->Height, 0.1f, 100.0f);

	glMatrixMode(GL_MODELVIEW); // Select the modelview matrix
	glLoadIdentity(); // Reset the modelview matrix

}
// ---------------------------------------------------------------------------
void __fastcall TForm2::OpenGLAPPanel1Paint(TObject *Sender)
{
	OffScreenRender();

}
// ---------------------------------------------------------------------------
void __fastcall TForm2::TimerPaintTimer(TObject *Sender)
{
	OffScreenRender();
}

// ---------------------------------------------------------------------------
void __fastcall TForm2::FormDestroy(TObject *Sender)
{
	delete Bitmap;
	Bitmap = NULL;
}

// ---------------------------------------------------------------------------
void __fastcall TForm2::OpenGLAPPanel1Init(TObject *Sender)
{
	glShadeModel(GL_SMOOTH); // Enable smooth shading
	glClearColor(0.0f, 0.0f, 0.0f, 0.5f); // Black background
	glClearDepth(1.0f); // Depth buffer setup
	glEnable(GL_DEPTH_TEST); // Enables depth testing
	glDepthFunc(GL_LEQUAL); // The type of depth testing to do
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	// Really nice perspective calculations
	TimerPaint->Enabled = true;
}
// ---------------------------------------------------------------------------
